/* This file is part of screensaver-settings
   Copyright (C) 2008 xFX JumpStart
 
   screensaver-settings is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 3 of the
   License, or (at your option) any later version.
 
   screensaver-settings is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
   General Public License for more details.

   You should have received a copy of the GNU Lesser General Public License
   along with screensaver-settings; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
   MA 02110-1301 USA */

using System;
using Gtk;
using System.IO;
using System.Collections.Generic;
using System.Diagnostics;
using Gnome;

public partial class MainWindow: Gtk.Window {
	static string GCONF_APP_PATH = "/apps/gnome-screensaver";
	static string IDLEDELAY_KEY = GCONF_APP_PATH + "/idle_delay";
	static string IDLEACTIVATION_KEY = GCONF_APP_PATH + "/idle_activation_enabled";
	static string LOCKSCREEN_KEY = GCONF_APP_PATH + "/lock_enabled";
	static string MODE_KEY = GCONF_APP_PATH + "/mode";
	static string THEMES_KEY = GCONF_APP_PATH + "/themes";
	static string RNDTHEMES_KEY = GCONF_APP_PATH + "/random_themes";
	static string CYCLEDELAY_KEY = GCONF_APP_PATH + "/cycle_delay";
	
	enum OperationModeConstants {
		BlankScreen = 0,
		OneScreenSaver = 1,
		Random = 2		
	}
	private OperationModeConstants opMode;
	
	private List<ScreenSaver> ScreenSavers = new List<ScreenSaver>();
	private List<string> rndSelScreensavers;
	private ScreenSaver selectedScreenSaver;
	private ulong previewWindowID = 0;
	private Process previewProcess;
	private SettingsUI settingsWin;
	private FullscreenPreviewWindow FullScreenWin;
	private GConf.Client client;
	
	public MainWindow () : base (Gtk.WindowType.Toplevel) {
		this.Build();
		
		btnClose.Clicked += CloseApp;
		btnSettings.Clicked += ShowSettingsWindow;
		btnFullscreen.Clicked += ToggleFullScreen;
		btnPowerManagement.Clicked += OpenPowerManagement;
		btnHelp.Clicked += ShowHelp;
		
		this.DeleteEvent += TerminateApp;
		
		//this.Build() issues this.Show(); thus, we're already visible by the
		//time we hit here.
		Init();
	}
	
	private void TerminateApp(object sender, DeleteEventArgs a) {
		CloseApp(this, new EventArgs());
	}
	
	private void LoadScreenSavers(bool showCheckBoxes) {
		int colNameID = (showCheckBoxes?1:0);
		
		// hmmm.. where's the tvScreenSavers.Columns.Clear() ???
		TreeViewColumn[] c = tvScreenSavers.Columns;
		for(int i = 0; i < c.Length; i++) {
			tvScreenSavers.RemoveColumn(c[i]);
		}
		
		ListStore ssStore = showCheckBoxes ? new ListStore(typeof(bool), typeof(string)) : new ListStore(typeof(string)); 
		tvScreenSavers.Model = ssStore;

		if(showCheckBoxes) {
			CellRendererToggle crToggle = new CellRendererToggle();
			crToggle.Activatable = true;
			crToggle.Toggled += SSToggled;
			tvScreenSavers.AppendColumn("CheckState", crToggle, "active", 0);
		}
		
		tvScreenSavers.HeadersVisible = false;
		tvScreenSavers.AppendColumn("Name", new CellRendererText(), "text", colNameID);
		tvScreenSavers.Selection.Changed += ScreenSaverSelected;
		tvScreenSavers.Selection.Mode = SelectionMode.Single;
		//tvScreenSavers.RowSeparatorFunc = new TreeViewRowSeparatorFunc(TreeViewSeparator, null);
	
		ssStore.SetSortFunc(colNameID, new TreeIterCompareFunc(SortList));
		ssStore.SetSortColumnId(colNameID, SortType.Ascending);
		
		ScreenSavers.Clear();
		foreach(string fileName in Directory.GetFiles(MainClass.SS_THEME_PATH, "*.desktop")) {
			DesktopItem theme;
			
			FileInfo f = new FileInfo(fileName);
			string userTheme = System.IO.Path.Combine(ScreenSaver.UserDesktopPath, f.Name);
			if(File.Exists(userTheme))
				theme = DesktopItem.NewFromFile(userTheme, DesktopItemLoadFlags.NoTranslations);
			else
				theme = DesktopItem.NewFromFile(fileName, DesktopItemLoadFlags.NoTranslations);
			
			if(theme == null) { //TODO: something better than silently fail to load theme
				continue;
			}
			TextReader configData;
			if(File.Exists(ScreenSaver.GetConfigFile(theme))) {
				configData = File.OpenText(ScreenSaver.GetConfigFile (theme));
			} else {
				configData = null;
			}
			ScreenSaver ss = new ScreenSaver(theme, configData);
			if(ss.IsValid) {
				ScreenSavers.Add(ss);
				if(showCheckBoxes) {
					ssStore.AppendValues(rndSelScreensavers.Contains(ss.Title), ss.Title);
				} else {					
					ssStore.AppendValues(ss.Title);
				}
			}
		}
		
		tvScreenSavers.ColumnsAutosize();
		if(ScreenSavers.Count > 0) {
			//This seems to prevent Xgl from crashing on my machine.
			System.Threading.Thread.Sleep(1000);
			
			//FIXME: I just can't belive this is correct (although it works).		
			string systemSelectedScreensaver = GetScreensaverFromGConf(false);
			foreach(ScreenSaver s in ScreenSavers) {
				if(s.Name == systemSelectedScreensaver) {
					systemSelectedScreensaver = s.Title;
					break;
				}
			}
			bool forceSelectFirst = true;
			TreeIter iter;
			for(int i = 0; i < ssStore.IterNChildren(); i++) {
				ssStore.GetIter(out iter, new TreePath(i.ToString()));
				if((string)ssStore.GetValue(iter, colNameID) == systemSelectedScreensaver) {
					TreePath iterPath = ssStore.GetPath(iter);
					tvScreenSavers.Selection.SelectPath(iterPath);
					tvScreenSavers.ScrollToCell(iterPath, tvScreenSavers.Columns[colNameID], false, 0, 0);
					forceSelectFirst = false;
					break;
				}
			}
			if(forceSelectFirst) {
				tvScreenSavers.Selection.SelectPath(new TreePath("0"));
			}
		} else {
			SetControlsState(false, opMode);
		}
	}
	
	private bool TreeViewSeparator(TreeModel model, TreeIter iter) {
		int colNameID = tvScreenSavers.Columns.Length - 1;
		
		return (string)model.GetValue(iter, colNameID) == "---";
	}
	
	private int SortList(TreeModel model, TreeIter iter1, TreeIter iter2) {
		int colNameID = tvScreenSavers.Columns.Length - 1;
		
		string title1 = (string)model.GetValue(iter1, colNameID);
		string title2 = (string)model.GetValue(iter2, colNameID);
				
		return title1.CompareTo(title2);
	}
	
	private void SSToggled(object sender, ToggledArgs e) {
		TreeIter iter;
		TreeModel model = tvScreenSavers.Model;
		
		if(model.GetIter(out iter, new TreePath(e.Path))) {
			bool currentState = (bool)model.GetValue(iter, 0);
			model.SetValue(iter, 0, !currentState);
			UpdateList((string)model.GetValue(iter, 1), !currentState);
		}
	}
	
	private void UpdateList(string title, bool state) {
		if(state) {
			if(!rndSelScreensavers.Contains(title)) rndSelScreensavers.Add(title);
		} else {
			rndSelScreensavers.Remove(title);
		}

		// FIXME: This is called every time the user checks or un-checks a screensaver
		// we should probably find a more suitable trigger to update the random list
		if(rndSelScreensavers.Count > 0) {
			client.Set(RNDTHEMES_KEY, rndSelScreensavers.ToArray());
		} else {
			client.Set(RNDTHEMES_KEY, new string[] {""});
		}
		UpdateSelectedScreensaver("");
	}
	
	private void ScreenSaverSelected(object sender, EventArgs e) {
		TreeIter iter;
		TreeModel model;
		TreeSelection sel = (TreeSelection)sender;
		int colNameID = tvScreenSavers.Columns.Length - 1;
				
		if(sel.GetSelected(out model, out iter)) {	
			string ssTitle = (string)model.GetValue(iter, colNameID);
				
			if(selectedScreenSaver != null && selectedScreenSaver.HasChanged) {
				selectedScreenSaver.ApplyCommandLine();
			}
						
			StopPreview();
			//HideSettingsWindow();
			UpdateSelectedScreensaver(ssTitle);
				
			selectedScreenSaver = null;
			btnSettings.Sensitive = (opMode != OperationModeConstants.BlankScreen);
			btnFullscreen.Sensitive = (opMode != OperationModeConstants.BlankScreen);

			if(opMode != OperationModeConstants.BlankScreen) {
				foreach(ScreenSaver ss in ScreenSavers) {
					if(ss.Title == ssTitle) {
						selectedScreenSaver = ss;
						StartPreview();
						if(settingsWin.GdkWindow == null) CreateSettingsWindow();
						if(settingsWin.GdkWindow.IsVisible) ShowSSSettings();
						UpdateSelectedScreensaver(selectedScreenSaver.Name);
					
						btnSettings.Sensitive = true;
						btnFullscreen.Sensitive = true;
					
						return;
					}
				}
			}
		}
	}

	private void ShowSettingsWindow(object sender, EventArgs e) {
		ShowSSSettings();
	}
	
	public void HideSettingsWindow() {
		if(settingsWin != null) settingsWin.Hide();
	}
	
	private void ToggleFullScreen(object sender, EventArgs e) {
		StopPreview();
		HideSettingsWindow();
		
		FullScreenWin = new FullscreenPreviewWindow();
		previewWindowID = MainClass.GetPreviewWindowID (FullScreenWin.PreviewArea);
		FullScreenWin.Init(this);
		FullScreenWin.Modal = true;
		FullScreenWin.Fullscreen();
		FullScreenWin.Destroyed += ExitedFullscreen;
		
		StartPreview();
	}
	
	private void ExitedFullscreen(object sender, EventArgs e) {
		FullScreenWin = null;
		StopPreview();
		
		previewWindowID = MainClass.GetPreviewWindowID(ssPreviewArea);
		StartPreview();
	}
	
	public void ShowSSSettings() {
		if(FullScreenWin != null) FullScreenWin.Destroy();

		settingsWin.Show();
		settingsWin.CreateUI(selectedScreenSaver);
	}
	
	private void OpenPowerManagement(object sender, EventArgs e) {
		Process p = new Process();
		p.StartInfo.FileName = "/usr/bin/gnome-power-preferences";
		p.Start();
		
		//p.WaitForExit();
	}
	
	private void ShowHelp(object sender, EventArgs e) {
		string msg = "";
		/*
		string readmeFile = String.Format("{0}README", System.Reflection.Assembly.GetExecutingAssembly().Location.Replace("screensaver-settings.exe", ""));
		FileInfo file = new FileInfo(readmeFile);
		if(file.Exists) {
			StreamReader sr = file.OpenText();
			msg = Environment.NewLine + sr.ReadToEnd().Replace("<", "&lt;");
			sr.Close();
		}settingsWin.Hide();
		*/
		Dialog dlg = new MessageDialog(this, 
		                               DialogFlags.DestroyWithParent, 
		                               MessageType.Info, 
		                               ButtonsType.Close, 
		                               "This feature has not been implemented yet" + msg);
		dlg.Run();
		dlg.Destroy();
	}
	
	private void CloseApp(object sender, EventArgs e) {
		this.Sensitive = false;
		StopPreview();	
		Application.Quit();
	}
	
	private void StartPreview() {
		StopPreview();
		try {
			previewProcess = new Process();
			
			//Console.WriteLine();
			//Console.WriteLine(String.Format("{0} {1} 0x{2:X}", selectedScreenSaver.CommandLineOptions, MainClass.WIN_REDIR, previewWindowID));
			//Console.WriteLine();

			previewProcess.StartInfo.FileName = selectedScreenSaver.Binary;
			previewProcess.StartInfo.Arguments = String.Format("{0} {1} {2}", selectedScreenSaver.CommandLineOptions, MainClass.WIN_REDIR, previewWindowID);
			previewProcess.Start();
			
			// This should be the way to go but, besides that it doesn't work, this function
			//	does not return a reference to the process ID, in the way the "gdk_spawn_on_screen_with_pipes"
			//	function does.
			//	Unfortunately this causes that some screensavers will not be displayed in the preview window.
			//Gdk.Spawn.CommandLineOnScreen(ssPreviewArea.GdkWindow.Screen, selectedScreenSaver.FileName + " " + selectedScreenSaver.CommandLine);
				
			if(FullScreenWin != null) {
				FullScreenWin.ScreensaverTitle = selectedScreenSaver.Title;
			}
		} catch(Exception e) {Console.WriteLine(e.Message);}
	}
	
	private void StopPreview() {
		//Console.WriteLine("Stopping Screensaver Preview");
		
		ssPreviewArea.GdkWindow.Clear();
		ssPreviewArea.GdkWindow.Background = Style.Black;

		if(previewProcess != null) {
			if(!previewProcess.HasExited) {
				try {
					previewProcess.Kill();
				} catch(Exception e) {Console.WriteLine(e.Message);}
			}
			previewProcess = null;
		}
	}
	
	public void SelectNextSS() {
		TreeIter iter;
		TreeModel model;
		TreeSelection sel = tvScreenSavers.Selection;
		
		if(sel.GetSelected(out model, out iter)) {
			int selIndex = model.GetPath(iter).Indices[0];
			tvScreenSavers.Selection.SelectPath(new TreePath((selIndex+1).ToString()));
		}
	}
	
	public void SelectPreviousSS() {
		TreeIter iter;
		TreeModel model;
		TreeSelection sel = tvScreenSavers.Selection;
		
		if(sel.GetSelected(out model, out iter)) {
			int selIndex = model.GetPath(iter).Indices[0];
			tvScreenSavers.Selection.SelectPath(new TreePath((selIndex-1).ToString()));
		}
	}
	
	private void HandleControlsChanged(object sender, EventArgs e) {
		UpdateControls();
	}
	
	private void UpdateControls() {
		OperationModeConstants mode = (OperationModeConstants)cmbMode.Active;
		SetControlsState(chkActivateSS.Active, mode);
		
		if(opMode == mode) return;
		opMode = mode;
		
		switch(opMode) {
		case OperationModeConstants.BlankScreen:
			StopPreview();
			break;
		case OperationModeConstants.OneScreenSaver:
			LoadScreenSavers(false);
			break;
		case OperationModeConstants.Random:
			LoadScreenSavers(true);
			break;
		}
	}
	
	private void SetControlsState(bool state, OperationModeConstants mode) {
		bool forceDisabled = (mode == OperationModeConstants.BlankScreen);
		
		hsIdleDelay.Sensitive = state;
		chkLockScreen.Sensitive = state;
		hbModeControls.Sensitive = state;
		hbRandomControls.Sensitive = state;
		hbRandomControls.Visible = (mode == OperationModeConstants.Random);
		swScreensavers.Sensitive = (state ? !forceDisabled : state);
		hbSSButtons.Sensitive = state;
		if(!state) {
			StopPreview();
		} else if(opMode != OperationModeConstants.BlankScreen) {
			StartPreview();
		}
	}

	private void Init() {
		previewWindowID = MainClass.GetPreviewWindowID(ssPreviewArea);
		CreateSettingsWindow();
		
		this.BorderWidth = (uint)10;
		
		client = new GConf.Client();
		GetSettingsFromGConf();
		client.AddNotify(GCONF_APP_PATH, new GConf.NotifyEventHandler(GConfSettingsChanged));
		
		hsCycleDelay.SetRange(1, 120);
		
		GetScreensaverFromGConf(true);
		cmbMode.Active = (int)opMode;
		LoadRandomScreenSaversList();
		LoadScreenSavers(opMode == OperationModeConstants.Random);
		UpdateControls();
		
		hsIdleDelay.ValueChanged += SaveSettingsToGConf;
		chkActivateSS.Toggled += SaveSettingsToGConf;
		chkActivateSS.Toggled += HandleControlsChanged;
		chkLockScreen.Toggled += SaveSettingsToGConf;
		cmbMode.Changed += HandleControlsChanged;
		cmbMode.Changed += SaveSettingsToGConf;
		hsCycleDelay.ValueChanged += SaveSettingsToGConf;
	}
	
	private void LoadRandomScreenSaversList() {
		rndSelScreensavers = new List<string>();
			
		try {
			string[] rndThemes = (string[])client.Get(RNDTHEMES_KEY);
			rndSelScreensavers.AddRange(rndThemes);
		} catch {
			foreach(ScreenSaver s in ScreenSavers) {
				rndSelScreensavers.Add(s.Title);
			}
		}
	}
		
	private void CreateSettingsWindow() {
		settingsWin = new SettingsUI();
		settingsWin.Hide();
		settingsWin.SettingsChanged += StartPreview;
		settingsWin.WindowPosition = Gtk.WindowPosition.Center;
	}
	
	private void GetSettingsFromGConf() {
		try {
			hsIdleDelay.Value = (int)client.Get(IDLEDELAY_KEY);
		} catch(Exception e) {Console.WriteLine(IDLEDELAY_KEY + " " + e.Message);}
		
		try {
			chkActivateSS.Active = (bool)client.Get(IDLEACTIVATION_KEY);
		} catch(Exception e) {Console.WriteLine(IDLEACTIVATION_KEY + " " + e.Message);}
		
		try {
			chkLockScreen.Active = (bool)client.Get(LOCKSCREEN_KEY);
		} catch(Exception e) {Console.WriteLine(LOCKSCREEN_KEY + " " + e.Message);}
		
		try {
			int value = 0;
			int.TryParse(client.Get(CYCLEDELAY_KEY).ToString(), out value);
		} catch(Exception e) {Console.WriteLine(CYCLEDELAY_KEY + " " + e.Message);}
	}
	
	private void SaveSettingsToGConf(object sender, EventArgs e) {
		client.Set(IDLEDELAY_KEY, (int)hsIdleDelay.Value);
		client.Set(IDLEACTIVATION_KEY, chkActivateSS.Active);
		client.Set(LOCKSCREEN_KEY, chkLockScreen.Active);
		
		switch(opMode) {
		case OperationModeConstants.BlankScreen:
			client.Set(MODE_KEY, "blank-only");
			break;
		case OperationModeConstants.Random:
			client.Set(CYCLEDELAY_KEY, hsCycleDelay.Value);
			client.Set(MODE_KEY, "random");
			break;
		default:
			client.Set(MODE_KEY, "single");
			break;
		}
	}
	
	private void UpdateSelectedScreensaver(string name) {		
		switch(opMode) {
		case OperationModeConstants.BlankScreen:
			client.Set(THEMES_KEY, new string[] {""});
			break;
		case OperationModeConstants.Random:
			string[] names;
			try {
				// Get the user's selected set of screensavers
				names = (string[])client.Get(RNDTHEMES_KEY);
				for(int i = 0; i < names.Length; i++) {
					foreach(ScreenSaver s in ScreenSavers) {
						if(s.Title == names[i]) names[i] = "screensavers-" + s.Name;
					}
				}
			} catch {
				// The key didn't exists so let's add them all...
				names = new string[ScreenSavers.Count];
				for(int i = 0; i < ScreenSavers.Count; i++) {
					names[i] = "screensavers-" + ScreenSavers[i].Name;
				}
			}
			
			client.Set(THEMES_KEY, names);
			break;
		default:
			client.Set(THEMES_KEY, new string[] {"screensavers-" + name});
			break;
		}
	}
	
	private string GetScreensaverFromGConf(bool isFirstTime) {
		string name = (string)client.Get(MODE_KEY);
		switch(name) {
		case "blank-only":
			if(isFirstTime) opMode = OperationModeConstants.BlankScreen;
			return "Blank Screen";
		case "random":
			if(isFirstTime) opMode = OperationModeConstants.Random;
			return "Random";
		}
		
		try {
			name = ((string[])client.Get(THEMES_KEY))[0];
			if(isFirstTime) opMode = OperationModeConstants.OneScreenSaver;
			return name.Remove(0, "screensavers-".Length);
		} catch {
			if(isFirstTime) opMode = OperationModeConstants.BlankScreen;
			return "blank-only";
		}
	}
			
	public void GConfSettingsChanged(object sender, GConf.NotifyEventArgs args) {
		GetSettingsFromGConf();
	}
	
	/*
	 This will be added as soon as the ALL/NONE/INV buttons are added to the GUI
	 
	 private void ApplySelection(object sender, EventArgs e) {
		const int SelAll = 0;
		const int SelNone = 1;
		const int SelInv = 2;
		
		int mode = SelAll;
		switch(((Button)sender).Name) {
		case "btnAll":	mode = SelAll;	break;
		case "btnNone":	mode = SelNone;	break;
		case "btnInv":	mode = SelInv;	break;
		}
		
		TreeIter iter;
		TreeModel model = tvScreenSavers.Model;
		bool state = false;
		
		for(int i = 0; i < tvScreenSavers.Model.IterNChildren(); i++) {
			model.GetIter(out iter, new TreePath(i.ToString()));
			switch(mode) {
				case SelAll:	state = true; break;
				case SelNone:	state = false; break;
				case SelInv:	state = !(bool)model.GetValue(iter, 0); break;
			}			
			model.SetValue(iter, 0, state);
			UpdateList((string)model.GetValue(iter, 1), state);
		}
	}
	*/ 
}