/* This file is part of screensaver-settings
   Copyright © 2008 Chris Halse Rogers <raof@ubuntu.com>
 
   screensaver-settings is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 3 of the
   License, or (at your option) any later version.
 
   screensaver-settings is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
   General Public License for more details.

   You should have received a copy of the GNU Lesser General Public License
   along with screensaver-settings; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
   MA 02110-1301 USA */

using System;
using System.IO;
using NUnit.Framework;
using Gnome;

namespace ScreensaverSettings
{
	[TestFixture()]
	public class TestScreensaver
	{
		private DesktopItem desktopFile;
		static string xmlHeader = "<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>";
		private TextReader emptyXml;
		
		private string tmpDir;
		private string MkTmpDir ()
		{
			if (String.IsNullOrEmpty (tmpDir)) {
				System.Diagnostics.ProcessStartInfo mktmpInfo = new System.Diagnostics.ProcessStartInfo ("mktemp", "-d");
				mktmpInfo.RedirectStandardOutput = true;
				mktmpInfo.UseShellExecute = false;
				                                                                                     
				System.Diagnostics.Process mktmp = System.Diagnostics.Process.Start(mktmpInfo);
				StreamReader asciiStream = new StreamReader (mktmp.StandardOutput.BaseStream, System.Text.Encoding.ASCII);
				tmpDir = asciiStream.ReadLine ();
			}
			if (String.IsNullOrEmpty (tmpDir)) {
				throw new ApplicationException ("Failed to read temp dir from process");
			}
			if (!Directory.Exists (tmpDir)) {
				string tmpError = tmpDir;
				tmpDir = null;
				throw new ApplicationException (String.Format ("Failed to create temp directory \"{0}\" for test",
				                                               tmpError));
			}
			return tmpDir;
		}
		
		[SetUp()]
		public void SetUp ()
		{
			Gtk.Application.Init ();
			tmpDir = null;
			desktopFile = new DesktopItem ();
			desktopFile.LocationFile = "mocksaver.desktop";
			desktopFile.SetString ("Name", "Mock Screensaver");
			desktopFile.SetString ("Exec", "mocksaver -root");
			emptyXml = new StringReader (xmlHeader + "<screensaver />");
		}
		
		[TearDown ()]
		public void TearDown ()
		{
			if (tmpDir != null) {
				Directory.Delete (tmpDir, true);
			}
		}
		[Test()]
		public void TestNameFromDesktop()
		{
			ScreenSaver saver = new ScreenSaver (desktopFile, emptyXml);
			Assert.AreEqual ("Mock Screensaver", saver.Name);
		}

		[Test()]
		public void TestCurrentSettingsFromDesktop ()
		{
			desktopFile.SetString ("Exec", "mocksaver -root -bar");
			ScreenSaver saver = new ScreenSaver (desktopFile, emptyXml);
			Assert.AreEqual ("mocksaver -bar", saver.CommandLine);
		}
		
		[Test()]
		public void TestCurrentSettingsFromDesktopSubstring ()
		{
			desktopFile.SetString ("Exec", "mocksaver -rootisnot -root");
			ScreenSaver saver = new ScreenSaver (desktopFile, emptyXml);
			Assert.AreEqual ("mocksaver -rootisnot", saver.CommandLine);
		}
		
		[Test()]
		public void TestGetConfigInfoFromDesktop ()
		{
			Assert.AreEqual ("/usr/share/xscreensaver/config/mocksaver.xml", 
			                 ScreenSaver.GetConfigFile (desktopFile));
		}
		
		[Test()]
		public void TestGetDescriptionFromXmlStream ()
		{
			string desc = "Draws a randomly-coloured Voronoi tessellation";
			ScreenSaver saver = new ScreenSaver (desktopFile,
			                                     new StringReader (xmlHeader + 
			                                                       String.Format ("<screensaver><_description>{0}</_description></screensaver>",
			                                                                      desc)));
			Assert.AreEqual (desc, saver.Description);
		}

		[Test()]
		public void TestGetCurrentSettingsIntFromDesktop ()
		{
			desktopFile.SetString ("Exec", "mocksaver -myopt 21 -root");
			ScreenSaver saver = new ScreenSaver (desktopFile, emptyXml);
			Assert.AreEqual ("21", saver.CurrentValues["-myopt"]);
		}

		[Test()]
		public void TestGetCurrentSettingsNegativeIntFromDesktop ()
		{
			desktopFile.SetString ("Exec", "mocksaver -myopt -20 -root");
			ScreenSaver saver = new ScreenSaver (desktopFile, emptyXml);
			Assert.AreEqual ("-20", saver.CurrentValues["-myopt"]);
		}

		[Test()]
		public void TestGetCurrentSettingsNoArgFromDesktop ()
		{
			desktopFile.SetString ("Exec", "mocksaver -myopt -root");
			ScreenSaver saver = new ScreenSaver (desktopFile, emptyXml);
			Assert.AreEqual ("", saver.CurrentValues["-myopt"]);
		}
		
		[Test()]
		public void TestScreensaverWithNullConfigHasNoConfig ()
		{
			ScreenSaver saver = new ScreenSaver (desktopFile, null);
			Assert.IsFalse (saver.Configurable);
		}
		
		[Test()]
		public void TestScreensaverWithValidConfigIsConfigurable ()
		{
			string desc = "<_description>Just a screensaver</_description>";
			string screenSaver = String.Format ("<screensaver>{0}</screensaver>", desc);
			
			StringReader configData = new StringReader (xmlHeader + screenSaver);
			ScreenSaver saver = new ScreenSaver (desktopFile, configData);
			Assert.IsTrue (saver.Configurable);
		}
		
		[Test()]
		public void TestGetBinaryDefaultPath ()
		{
			desktopFile.SetString ("Exec", "saverexec -someopt otheropt");
			ScreenSaver saver = new ScreenSaver (desktopFile, null);
			Assert.AreEqual (Path.Combine (MainClass.SS_BIN_PATH, "saverexec"),
			                 saver.Binary);
		}
		
		[Test()]
		public void TestGetCommandOptions ()
		{
			desktopFile.SetString ("Exec", "somebin -optone -opttwo 29");
			ScreenSaver saver = new ScreenSaver (desktopFile, null);
			Assert.AreEqual ("-optone -opttwo 29",
			                 saver.CommandLineOptions);
		}
		
		[Test()]
		public void SaveDesktopToPath ()
		{
			desktopFile.SetString ("Name", "testname");
			ScreenSaver saver = new ScreenSaver (desktopFile, null);
			string tmpSavePath = MkTmpDir ();
			saver.SaveConfig (tmpSavePath);
			Assert.IsTrue (File.Exists (Path.Combine (tmpSavePath,
			                                          saver.ConfigFileName)));
		}
		
		[Test()]
		public void SavedConfigHasCorrectContents ()
		{
			desktopFile.SetString ("Name", "testname");
			desktopFile.SetString ("SomethingElse", "just a value");
			ScreenSaver saver = new ScreenSaver (desktopFile, null);
			string tmpSavePath = MkTmpDir ();
			saver.SaveConfig (tmpSavePath);
			DesktopItem savedFile = DesktopItem.NewFromFile (Path.Combine (tmpSavePath,
			                                                               saver.ConfigFileName), DesktopItemLoadFlags.NoTranslations);
			Assert.AreEqual ("testname", savedFile.GetString ("Name"));
			Assert.AreEqual ("just a value", savedFile.GetString ("SomethingElse"));
		}
		
		[Test()]
		public void SavedConfigUsesRootOption ()
		{
			desktopFile.SetString ("Exec", "executable -options -that -do -not -include -hyphenroot");
			ScreenSaver saver = new ScreenSaver (desktopFile, null);
			string tmpSavePath = MkTmpDir ();
			saver.SaveConfig (tmpSavePath);
			DesktopItem savedFile = DesktopItem.NewFromFile (Path.Combine (tmpSavePath,
			                                                               saver.ConfigFileName), DesktopItemLoadFlags.NoTranslations);
			Assert.IsTrue (savedFile.GetString ("Exec").Contains("-root"));
		}
		
		[Test()]
		public void SaveConfigDoesNotChangeCommandLine ()
		{
			desktopFile.SetString ("Exec", "screensaver -someopt -notroot");
			ScreenSaver saver = new ScreenSaver (desktopFile, null);
			string oldCommandLine = saver.CommandLine;
			saver.SaveConfig (MkTmpDir ());
			Assert.AreEqual (oldCommandLine, saver.CommandLine);
		}
		
		[Test()]
		[ExpectedExceptionAttribute ("System.ApplicationException", "Unable to save config to \"/path/which/does/not/exist\"")]
		public void SaveConfigRaisesWhenUnableToWrite ()
		{
			ScreenSaver saver = new ScreenSaver (desktopFile, null);
			saver.SaveConfig ("/path/which/does/not/exist");
		}
		
		[Test()]
		public void SavedConfigHasSameFilename ()
		{
			desktopFile.LocationFile = "myconfig.desktop";
			ScreenSaver saver = new ScreenSaver (desktopFile, null);
			
			string tmpSavePath = MkTmpDir ();
			saver.SaveConfig (tmpSavePath);
			Assert.IsTrue (File.Exists (Path.Combine (tmpSavePath, "myconfig.desktop")));
		}
		
		[Test()]
		public void GetDesktopFileName ()
		{
			desktopFile.LocationFile = "myconfig.desktop";
			ScreenSaver saver = new ScreenSaver (desktopFile, null);
			Assert.AreEqual ("myconfig.desktop", saver.ConfigFileName);
		}
	}
}
